from lextok.rules._pattern import Label, Rule, _orth_in

prefix_tits = _orth_in(["Atty.", "Hon.", "Engr.", "Dr.", "Dra."])
prefix_titled_person = Rule(
    label=Label.PERSON,
    patterns=[
        [prefix_tits, {"IS_TITLE": True, "OP": "+"}],
        [prefix_tits, {"ENT_TYPE": Label.PERSON.name}],
    ],
)



from typing import Any

from lextok.rules._pattern import (
    CM,
    OF,
    Label,
    Rule,
    _orth_in,
    _re,
    create_regex_options,
)
from lextok.rules.abbreviations import Abbv

months = [
    month
    for defined_month in (
        Abbv.January,
        Abbv.February,
        Abbv.March,
        Abbv.April,
        Abbv.May,
        Abbv.June,
        Abbv.July,
        Abbv.August,
        Abbv.Sept1,
        Abbv.Sept2,
        Abbv.October,
        Abbv.November,
        Abbv.December,
    )
    for month in defined_month.value.options
]

days = _orth_in([f"{str(i)}" for i in range(1, 32)] + [f"0{i}" for i in range(1, 10)])

ranged_years = create_regex_options(["(19\\d{2})", "(20\\d{2})"])

date_us: list[list[dict[str, Any]]] = [
    [CM | {"OP": "?"}, {"ORTH": month_name}, days, CM | {"OP": "?"}, _re(ranged_years)]
    for month_name in months  # , Feb. 01, 2023
]

date_uk: list[list[dict[str, Any]]] = [
    [CM | {"OP": "?"}, days, {"ORTH": month_name}, CM | {"OP": "?"}, _re(ranged_years)]
    for month_name in months  # 01 Feb 2023
]


covered_year: list[dict[str, Any]] = [
    _re(
        create_regex_options(["\\(", "\\["])
        + ranged_years  # [2023]
        + create_regex_options(["\\)", "\\]"])
    ),
]

series_year: list[dict[str, Any]] = [
    CM | {"OP": "?"},
    {"LOWER": {"IN": ["s.", "series"]}},  # , s. of 2023
    OF | {"OP": "?"},
    _re(ranged_years),
]

date_as_entity = Rule(
    label=Label.DATE,
    patterns=date_us + date_uk + [covered_year, series_year],
)



from typing import Any

from lextok.rules._pattern import OF, THE, Label, Rule
from lextok.rules.abbreviations import org_options

inc_suffixes = list(set(o.lower() for o in org_options)) and ["gmbh"]
inc_options: list[dict[str, Any]] = [
    {"ORTH": {"IN": [",", "and", "&"]}, "OP": "*"},
    {"LOWER": {"IN": inc_suffixes}},
    {"LOWER": {"IN": inc_suffixes + [","]}, "OP": "?"},
]
juridical_org = Rule(
    label=Label.JuridicalEntity,
    patterns=[
        [
            {
                "LOWER": {
                    "IN": (
                        "rep rep. republic people pp p.p. pp. govt govt. government"
                        " gov't".split()
                    )
                }
            },
            OF,
            THE,
            {"LOWER": {"IN": "phil phils phil. phils. philippines".split()}},
        ],
        [{"POS": "PROPN", "OP": "+"}] + inc_options,
        [{"IS_UPPER": True, "OP": "+"}] + inc_options,
        [{"IS_TITLE": True, "OP": "+"}] + inc_options,
        [{"ENT_TYPE": Label.ORG.name}] + inc_options,
        [{"ORTH": "Province"}, OF, {"IS_TITLE": True, "OP": "+"}],
        [{"ORTH": "City"}, OF, {"IS_TITLE": True, "OP": "+"}],
        [{"ORTH": "Municipality"}, OF, {"IS_TITLE": True, "OP": "+"}],
        [{"ORTH": "Estate"}, OF, {"IS_TITLE": True, "OP": "+"}],
        [{"ORTH": "Estate"}, OF, {"ENT_TYPE": Label.PERSON.name}],
    ],
)



import importlib
import io
import itertools
import re
import string
from collections.abc import Iterator
from enum import Enum, auto
from pathlib import Path
from typing import Any, NamedTuple

import inflect
import jsonlines
import roman  # type: ignore
from pydantic import BaseModel, model_serializer
from slugify import slugify
from spacy.tokens import Doc, Span
from spacy.util import filter_spans

camel_case_pattern = re.compile(
    r".+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)"
)


def uncamel(text: str) -> list[str]:
    """For text in camelCaseFormatting, convert into a list of strings."""
    return [m.group(0) for m in camel_case_pattern.finditer(text)]


class Label(Enum):
    """A `Label` is shorthand for creating a spacy `ENT_TYPE` for entity-based patterns
    and `LABEL`s for span patterns. An alternate route would be to use 2 types of
    data structures (one for entities and one for spans) but this adds complexity downstream
    since all `Rule`s will now have to have different label fields rather than just a single one.

    Each member of the `Label` Enum will contain properties and methods:

    1. `@node` to represent `{"ENT_TYPE": <Enum member name>}`
    2. `@opt`: adds `{"OP": "?}` to the `@node`
    3. `@snakecase`: each label gets converted to a snakecased variable that can be used as a custom Doc attribute
    4. `extract_entities(doc)`: get all Spans that are entities a _member_ `Label`
    5. `extract_spans(doc)`: : get all Spans that match a SpanRuler key which is a _member_ `Label`
    """

    PERSON = auto()
    NORP = auto()
    ORG = auto()
    DATE = auto()
    Actor = auto()
    JuridicalEntity = auto()
    GovtDivision = auto()
    CourtName = auto()
    Concept = auto()  # span
    Paper = auto()  # span
    ProvisionNum = auto()
    StatuteNum = auto()
    StatuteNamed = auto()
    StatutoryProvision = auto()  # span
    StatutoryLink = auto()  # span
    CaseName = auto()
    DocketNum = auto()
    ReporterNum = auto()
    DecisionCitation = auto()  # span
    GenericNum = auto()
    Candidate = auto()  # span
    CandidateProvision = auto()  # span

    @property
    def node(self) -> dict[str, str]:
        return {"ENT_TYPE": self.name}

    @property
    def opt(self) -> dict[str, str]:
        return self.node | {"OP": "?"}

    @property
    def snakecase(self) -> str:
        """Used as the variable for spacy's custom attributes e.g. `doc._.provision_nums`"""
        return "_".join(uncamel(self.name)).lower() + "s"

    def extract_entities(self, doc: Doc) -> list[Span]:
        """Generate doc's entities matching a label."""
        return [ent for ent in doc.ents if ent.label_ == self.name]

    def extract_spans(self, doc: Doc) -> list[Span]:
        """Generate doc's span ruler patterns matching a label."""
        return [span for span in filter_spans(doc.spans[self.name])]


SPAN_RULER_LABELS = [
    Label.Actor,
    Label.Concept,
    Label.Paper,
    Label.DecisionCitation,
    Label.StatutoryProvision,
    Label.StatutoryLink,
    Label.Candidate,
    Label.CandidateProvision,
]
"""When included in this list, consider the `Rule` associated with the `Label` a _SpanRuler_ pattern."""


excludables = [Label.DATE] + SPAN_RULER_LABELS
"""No need to include these in the custom attributes of each Doc."""

ENTITY_RULER_LABELS = [member for member in Label if member not in excludables]
"""When included in this list, consider the `Rule` associated with the `Label` an _EntityRuler_ pattern."""


class Rule(BaseModel):
    """`patterns` associated with a single `label` (optionally
    with an `id` as well that serve as the `ent_id_` in `spacy.tokens.Doc.ents`).
    It can also be used for pattern objects for `spacy.tokens.Doc.spans`. See generally:

    1. https://spacy.io/usage/rule-based-matching#entityruler-files
    2. https://spacy.io/usage/rule-based-matching#spanruler-files

    A `Rule` enables the creation of such pattern objects containing the same `Label` and custom `id`, if provided. Sample rule:

    ```py
    sample = Rule(
        id="ministry-labor",
        label=Label.GovtDivision,
        patterns=[
            [
                {"LOWER": "the", "OP": "?"},
                {"LOWER": "ministry"},
                {"LOWER": "of"},
                {"LOWER": "labor"},
            ]
        ],
    )
    ```

    Use with: <Rule instance>.create_file() or <Rule instance>.model_dump() to
    simply get a list of patterns without file creation. See common usage in
    `set_entity_ruler()`
    """

    id: str | None = None
    label: Label
    patterns: list[list[dict[str, Any]]]

    def __str__(self) -> str:
        return f"<Rule {self.ent_id})>"

    def __repr__(self) -> str:
        return f"<Rule {self.ent_id}: {len(self.patterns)} patterns>"

    @property
    def ent_id(self):
        return self.id or "-".join(uncamel(self.label.name)).lower()

    @model_serializer
    def ser_model(self) -> list[dict[str, Any]]:
        """Following the pydantic convention for .model_dump(); instead of a traditional
        `dict` return, the function results in a serialized list of patterns for consumption
        by either `create_file()` or the _entity_ruler_ spacy pipeline."""
        return [
            {"id": self.ent_id, "label": self.label.name, "pattern": pattern}
            for pattern in self.patterns
        ]

    def create_file(self, file: Path | None = None):
        """Will update the file, if it exists; will create a file, if it doesn't exist."""
        if not file:
            file = Path(__file__).parent.joinpath(f"{self.ent_id}.jsonl")
        fp = io.BytesIO()
        with jsonlines.Writer(fp) as writer:
            for ser_pattern in self.model_dump():
                writer.write(ser_pattern)
        file.unlink(missing_ok=True)
        file.write_bytes(fp.getbuffer())
        return file

    @classmethod
    def extract_from_files(
        cls,
        folder: Path,
        glob_pattern: str = "*.py",
        excluded: tuple = ("__init__", "_pattern"),
        ents: bool = True,
    ) -> Iterator["Rule"]:
        """This method of dynamically extracting from files is adopted from:
        https://stackoverflow.com/a/50181975/9081369

        1. Look for `Rule` instances within a `folder`
        2. Each `Rule` instance's will contain a `.label` value
        3. If the `.label` is `True`, this implies extracting _entity_ rules
        4. If the `.label` is `False`, this implies extracting _span_ rules

        Args:
            ents (bool, optional): See `Label` object whose value is `True` or `False`. Defaults to `True`.

        Yields:
            Iterator[Rule]: Rules that match the `ents` configuration.
        """
        for file in folder.glob(glob_pattern):
            if file.stem not in excluded:
                modules = {}
                spec = importlib.util.spec_from_file_location(file.stem, file)  # type: ignore
                modules[file.stem] = importlib.util.module_from_spec(spec)  # type: ignore
                spec.loader.exec_module(modules[file.stem])  # type: ignore
                for var_name, rule_candidate in modules[file.stem].__dict__.items():
                    if not var_name.startswith("__"):
                        if isinstance(rule_candidate, Rule):
                            if not rule_candidate.id:
                                rule_candidate.id = slugify(var_name)
                            if ents:
                                if rule_candidate.label not in SPAN_RULER_LABELS:
                                    yield rule_candidate
                            else:
                                if rule_candidate.label in SPAN_RULER_LABELS:
                                    yield rule_candidate


def _re(v: str, anchored: bool = True) -> dict[str, Any]:
    """Helper function to add an anchored, i.e. `^`<insert value `v` here>`$`
    regex pattern, following the convention `{"TEXT": {"REGEX": f"^{v}$"}}`
    spacy convention.
    """
    if anchored:
        v = f"^{v}$"
    return {"TEXT": {"REGEX": v}}


def _orth_in(v_list: list[str]):
    """Helper function to add a regex pattern where the options contained
    in the `v_list` are limited, following the convention `{"ORTH": {"IN": v_list}}`
    spacy convention.
    """
    return {"ORTH": {"IN": v_list}}


def _multi(pattern: list[dict[str, Any]], count: int) -> list[list[dict[str, Any]]]:
    """Create cumulative `pattern`s based on the `count` indicated.
    e.g. `repeat_pattern([1,2])` results in:

    [
        [1, 2],
        [1, 2, 1, 2],
        [1, 2, 1, 2, 1, 2],
        [1, 2, 1, 2, 1, 2, 1, 2],
        [1, 2, 1, 2, 1, 2, 1, 2, 1, 2]
    ]
    ```
    """
    res: list[list[dict[str, Any]]] = []
    res.append(pattern)
    for i in range(count):
        added = pattern + res[-1]
        res.append(added)
    return res


def create_regex_options(texts: list[str]):
    return "(" + "|".join(texts) + ")"


# REUSABLES


OF = {"LOWER": "of"}
"""Required `the`: `{"LOWER": "of"}`"""

TH = {"LOWER": "the", "OP": "?"}
"""Optional `the`: `{"LOWER": "the", "OP": "?"}`"""

COURT = {"ORTH": "Court"}
"""Required title cased `Court`"""

CODE = {"ORTH": "Code"}
"""Required title cased `Code`"""

THE = {"LOWER": "the"}
"""Required `the`: `{"LOWER": "the"}`"""

VS = _orth_in(["v.", "vs."])
"""Common formula for consistency: `{"ORTH": {"IN": ["v.", "vs."]}}`"""

PH = _orth_in(["Phil", "Phil.", "Phils", "Phils.", "Philippines"])

CM = {"ORTH": ","}

OF_THE_PH_ = [OF, THE, PH]

NUM_SYM = _orth_in(["No", "Nos", "No.", "Nos."])

OPT_NUMS = {"POS": {"IN": ["NUM", "PUNCT"]}, "OP": "*"}

CONNECTOR = _orth_in(["of", "the", ",", "and", "&"]) | {"OP": "*"}
"""Setting this optional token ("of", "the", ",", "and", "&") allows for
ProvisionNums/serials to be merged later"""

### PATTERN FUNCS


def lower_words(words: str):
    """Will separate each word separated by spaces into a
    `{"LOWER": <word.strip().lower()>}` spacy pattern."""
    return [{"LOWER": word.strip().lower()} for word in words.split()]


def lower_word_lists(word_lists: list[str]) -> list[list[dict[str, Any]]]:
    """Will separate each word separated by spaces into a
    `{"LOWER": <word.strip().lower()>}` spacy pattern."""
    return [lower_words(words) for words in word_lists]


def titled_words(words: str):
    """Will separate each word separated by spaces into a
    `{"LOWER": <word.lower()>, "IS_TITLE": True}` spacy pattern."""
    return [w | {"IS_TITLE": True} for w in lower_words(words)]


def name_code(fragments: str):
    """Use of `lower_words()` for Code names"""
    return [TH, {"LOWER": "code"}, OF] + lower_words(fragments)


def name_court(fragments: str):
    """Create title-based Court names"""
    return [TH, COURT, OF] + titled_words(fragments)


def name_statute(
    fragments: str | list[dict[str, Any]],
    options: list[str] = ["law", "act"],
) -> list[dict[str, Any]]:
    """If a string is passed as fragments, uses `lower_words()`
    in between options for named laws; if what is passed is a list of
    these same spacy patterns, these will be added as options."""
    bits: list[dict[str, Any]] = [TH]
    if isinstance(fragments, str):
        bits.extend(lower_words(fragments))
    elif isinstance(fragments, list):
        bits.extend(fragments)
    bits.append({"LOWER": {"IN": options}, "IS_TITLE": True})
    return bits


## DIGITS


p = inflect.engine()

roman_upper = [roman.toRoman(i) for i in range(1, 100)]
a_z_lower = [i for i in string.ascii_lowercase]
pairs = itertools.combinations_with_replacement(a_z_lower, r=2)
aa_zz_lower = [f"{a}{b}" for a, b in pairs]


class DigitLists(Enum):
    HundredDigit = [str(i) for i in range(0, 100)]
    WordHundredDigit = [p.number_to_words(num=i) for i in range(1, 100)]  # type: ignore
    RomanHundredLower = [i.lower() for i in roman_upper]
    RomanHundredUpper = [roman.toRoman(i) for i in range(1, 100)]
    AtoZSingleLower = a_z_lower
    AtoZSingleUpper = [i.upper() for i in a_z_lower]
    AtoZDoubleLower = aa_zz_lower
    AtoZDoubleUpper = [i.upper() for i in aa_zz_lower]

    @classmethod
    def generate_options(cls) -> list[str]:
        options: list[str] = []
        for member in cls:
            options.extend(member.value)  # type: ignore
        return options


HAS_DIGIT = _re(v=".*\\d.*")
"""Any token containing a digit should be used in tandem with an attribute ruler."""

DOTTED = _re(v="(\\d\\.)+\\d?")

IS_COVERED = _re(v=".*\\(\\w{1,2}\\).*")
"""Any token containing a digit should be used in tandem with an attribute ruler."""

SPECIFIC = _re("(" + "|".join(DigitLists.generate_options()) + ")")
"""Any token matching the options created by DigitLists"""

IS_ROMAN = _re(v="[IXV]+[-\\.][A-Z]{1,2}")
"""Handle combinations like I-A"""

PROV_DIGITS = [SPECIFIC, HAS_DIGIT, IS_COVERED, IS_ROMAN, DOTTED]


class Duo(NamedTuple):
    """Given two letters, create possible patterns using uppercase text."""

    a: str
    b: str

    @property
    def x(self):
        return self.a.upper()

    @property
    def y(self):
        return self.b.upper()

    @property
    def as_token(self) -> dict[str, list[dict[str, str]]]:
        """Used to create special rules for custom tokenizer."""
        return {f"{self.x}.{self.y}.": [{"ORTH": f"{self.x}.{self.y}."}]}

    @property
    def v1(self) -> list[dict[str, str]]:
        # R . A .
        return [{"ORTH": self.x}, {"ORTH": "."}, {"ORTH": self.y}, {"ORTH": "."}]

    @property
    def v2(self) -> list[dict[str, str]]:
        return [{"ORTH": f"{self.x}."}, {"ORTH": f"{self.y}."}]  # R. A.

    @property
    def v3(self) -> list[dict[str, str]]:
        return [{"ORTH": f"{self.x}.{self.y}."}]  # R.A.

    @property
    def v4(self) -> list[dict[str, str]]:
        return [{"ORTH": f"{self.x}{self.y}"}]  # RA

    @property
    def patterns(self) -> list[list[dict[str, str]]]:
        return [self.v1, self.v2, self.v3, self.v4]

    def add_to_each_pattern(self, terminators: list[dict[str, Any]]):
        for p in self.patterns:
            yield p + terminators


class Style(NamedTuple):
    """A `Style` refers to the way a specific two-`let`ter acronym can be used to create
    various letter patterns along with the word variants described in `v`.
    If no `let` exists, will only create word patterns using `v`."""

    v: list[str] = []
    let: str | None = None
    pre: list[str] = ["No", "Nos", "No.", "Nos."]
    r: str = "[\\w-]+"

    @property
    def title_pre(self) -> list[dict[str, Any]]:
        return [_orth_in(self.pre), _re(self.r)]

    @property
    def upper_pre(self) -> list[dict[str, Any]]:
        return [_orth_in([i.upper() for i in self.pre]), _re(self.r)]

    @property
    def token_parts(self):
        """The first pass is for indiscriminate words e.g. `bar matter`; the second, for
        dealing with periods, e.g. `adm. matter`. The first will generate the following as
        token parts ('bar','matter'); the second: ('adm.','matter'), ('adm','.','matter')
        """
        objs = set()
        for words in self.v:
            partial = []
            for word in words.split():
                partial.append(word)
            objs.add(tuple(partial))
        for words in self.v:
            partial = []
            for word in words.split():
                if word.endswith("."):
                    cleaned = word.removesuffix(".")
                    partial.append(cleaned)
                    partial.append(".")
                else:
                    partial.append(word)
            objs.add(tuple(partial))
        return objs

    @property
    def _title_num(self) -> list[list[dict[str, Any]]]:
        return [
            [{"ORTH": sub.title()} for sub in subtokens] + self.title_pre
            for subtokens in self.token_parts
        ]

    @property
    def _title_no_num(self) -> list[list[dict[str, Any]]]:
        return [
            [{"ORTH": sub.title()} for sub in subtokens] + [_re(self.r)]
            for subtokens in self.token_parts
        ]

    @property
    def _upper_num(self) -> list[list[dict[str, Any]]]:
        return [
            [{"ORTH": sub.upper()} for sub in subtokens] + self.upper_pre
            for subtokens in self.token_parts
        ]

    @property
    def _upper_no_num(self) -> list[list[dict[str, Any]]]:
        return [
            [{"ORTH": sub.upper()} for sub in subtokens] + [_re(self.r)]
            for subtokens in self.token_parts
        ]

    @property
    def initials(self) -> Duo | None:
        if not self.let:
            return None
        if len(self.let) != 2:
            return None
        return Duo(a=self.let[0], b=self.let[1])

    @property
    def word_patterns(self) -> list[list[dict[str, Any]]]:
        patterns = []
        patterns.extend(self._title_num)
        patterns.extend(self._title_no_num)
        patterns.extend(self._upper_num)
        patterns.extend(self._upper_no_num)
        return patterns

    @property
    def letter_patterns(self) -> list[list[dict[str, Any]]]:
        items: list[list[dict[str, Any]]] = []
        if not self.initials:
            return items
        for target_nodes in (self.upper_pre, self.title_pre, [_re(self.r)]):
            for b in self.initials.add_to_each_pattern(target_nodes):
                items.append(b)
        return items

    @property
    def patterns(self) -> list[list[dict[str, Any]]]:
        words = self.word_patterns
        letters = self.letter_patterns
        return words + letters if letters else words



import spacy
from spacy.lang.char_classes import (
    ALPHA,
    ALPHA_LOWER,
    ALPHA_UPPER,
    CONCAT_QUOTES,
    LIST_ELLIPSES,
    LIST_ICONS,
)

from lextok.rules.abbreviations import Abbv, Prov
from lextok.rules.rule_docketnums import DocketNum
from lextok.rules.rule_statutenums import StatuteNum

# Remove hyphen '-' as infix, see https://spacy.io/usage/linguistic-features#native-tokenizer-additions
INFIXES_OVERRIDE = (
    LIST_ELLIPSES
    + LIST_ICONS
    + [
        r"(?<=[0-9])[+\\-\\*^](?=[0-9-])",
        r"(?<=[{al}{q}])\\.(?=[{au}{q}])".format(
            al=ALPHA_LOWER, au=ALPHA_UPPER, q=CONCAT_QUOTES
        ),
        r"(?<=[{a}]),(?=[{a}])".format(a=ALPHA),
        # ✅ Commented out regex that splits on hyphens between letters:
        # r"(?<=[{a}])(?:{h})(?=[{a}])".format(a=ALPHA, h=HYPHENS),
        r"(?<=[{a}0-9])[:<>=/](?=[{a}])".format(a=ALPHA),
    ]
)

PAIRS = [("(", ")"), ("[", "]")]


def custom_prefix_list(nlp: spacy.language.Language):
    """Only use prefix `(` and `[` if _not_ followed by a single word `\\w+`
    with a closing `)` or `]`

    Note that modifications to a prefix should be done after the prefix removed, e.g.
    if the prefix is `(`, modify the `(`<add here> when appending a new rule.
    This is because of `compile_suffix_regex` which appends a `^` at the start of every prefix.

    The opening `open` will be considered a prefix only if not subsequently terminated by a closing `close`.

    example | status of `(`
    --:|:--
    `(`Juan de la Cruz v. Example) | is prefix
    Juan `(`de) la Cruz v. Example | is _not_ prefix
    """

    pfx = list(nlp.Defaults.prefixes)  # type: ignore
    for opened, closed in PAIRS:
        pfx.remove(f"\\{opened}")
        pfx.append(f"\\{opened}(?![\\w\\.]+\\{closed})")
    return pfx


def custom_suffix_list(nlp: spacy.language.Language):
    """Enable partner closing character, e.g. `)` or `]` to be excluded as a suffix
    if matched with an opening `(` or `[` within the range provided by `num`.

    Assuming a range of 20, this means 19 characters will be allowed:

    Let's exceed this range with 22, this results in a split of the terminal character `)`:

    ```py
    text = "twenty_two_char_string"
    len(text)  # 22
    nlp.tokenizer.explain("A (twenty_two_char_string)")
    # [('TOKEN', 'A'),
    # ('PREFIX', '('),
    # ('TOKEN', 'twenty_two_char_string'),
    # ('SUFFIX', ')')]
    ```

    This becomes the exception to the general rule that the closing suffix `)` should
    be removed from the custom tokenizer.

    However, if the number of characters within a closed / covered single world is 19 and below:

    ```py
    text = "smol"
    len(text)  # 4
    nlp.tokenizer.explain("A (smol)")
    # [('TOKEN', 'A'),
    # ('TOKEN', '(smol)'),
    # ('TOKEN', 'word')]
    ```

    The suffix ")" is removed per the general rule.
    """
    sfx = list(nlp.Defaults.suffixes)  # type: ignore
    for opened, closed in PAIRS:
        sfx.remove(f"\\{closed}")
        _pre = "".join([f"(?<!\\{opened}\\w{{{i}}})" for i in range(1, 20)])
        sfx.append(f"{_pre}\\{closed}")

    return sfx


def create_special_rules():
    def make_special_dotted(texts: list[str]):
        """Add a period after every text item in `texts`, to consider each a single token.
        These patterns can be used as a special rule in creating a custom tokenizer."""
        return {f"{t}.": [{"ORTH": f"{t}."}] for t in texts if not t.endswith(".")}

    x = make_special_dotted(
        "Rep Sen vs Vs v s et al etc Ll p Pp PP P.P R.P H.B S.B a.k.a".split()
    )

    a = {
        k: v
        for member in DocketNum
        if member.value.initials
        for k, v in member.value.initials.as_token.items()
    }

    b = {
        k: v
        for member in StatuteNum
        if member.value.initials
        for k, v in member.value.initials.as_token.items()
    }

    c = {
        f"{bit}.": [{"ORTH": f"{bit}."}]
        for style in (None, "lower", "upper")
        for bit in Abbv.set_abbvs(cased=style)
    }

    d = {
        f"{bit}.": [{"ORTH": f"{bit}."}]
        for style in (None, "lower", "upper")
        for bit in Prov.set_abbvs(cased=style)
    }

    return a | b | c | d | x



from enum import Enum

from lextok.rules._pattern import Label, Rule, Style


class StatuteNum(Enum):
    RA = Style(let="ra", r="\\d{1,5}", v=["republic act", "rep. act"])
    CA = Style(let="ca", r="\\d{1,3}", v=["commonwealth act", "com. act"])
    BP = Style(let="bp", r="\\d{1,3}(-?(A|B))?", v=["batas pambansa"], pre=["Blg."])
    EO = Style(let="eo", r="\\d{1,4}(-?(A|B|C))?", v=["executive order", "exec. order"])
    PD = Style(
        let="pd",
        r="\\d{1,4}(-?(A|B))?",
        v=["presidential decree", "pres. decree", "pres. dec."],
    )
    ACT = Style(r="\\d{1,4}", v=["act"])


statute_nums = Rule(
    label=Label.StatuteNum,
    patterns=[p for mem in StatuteNum for p in mem.value.patterns],
)



from enum import Enum

from lextok.rules._pattern import Label, Rule, Style


class DocketNum(Enum):
    GR = Style(let="gr")
    AM = Style(
        let="am",
        v=[
            "adm mat",
            "adm. mat",
            "adm mat.",
            "adm. mat.",
            "adm matter",
            "adm. matter",
            "admin. matter",
            "admin matter",
            "administrative matter",
        ],
    )
    AC = Style(
        let="ac",
        v=[
            "adm case",
            "adm. case",
            "admin. case",
            "admin case",
            "administrative case",
        ],
    )
    BM = Style(let="bm", v=["bar mat.", "bar matter"])


docket_nums = Rule(
    label=Label.DocketNum, patterns=[p for mem in DocketNum for p in mem.value.patterns]
)



from enum import Enum

from lextok.rules._pattern import COURT, Label, Rule, name_court, titled_words


class CourtName(Enum):
    SC = [{"LOWER": "supreme", "IS_TITLE": True}, COURT]
    RTC = titled_words("regional trial court")
    CA = name_court("appeals")
    CFI = name_court("first instance")
    MeTC = titled_words("metropolitan trial court")
    MCTC = titled_words("municipal circuit court")
    MTC = titled_words("municipal trial court")
    MTCC = (
        titled_words("municipal trial court")
        + [{"LOWER": "in"}]
        + titled_words("the cities")
    )


court_names = Rule(label=Label.CourtName, patterns=[mem.value for mem in CourtName])  # type: ignore



from lextok.rules._pattern import (
    COURT,
    OF,
    OF_THE_PH_,
    TH,
    Label,
    Rule,
    _orth_in,
    lower_words,
    titled_words,
)

generic_laws = Rule(
    label=Label.StatuteNamed,
    patterns=[
        [
            TH,
            _orth_in(["1987", "1973", "1935"]) | {"OP": "?"},
            _orth_in(["CONSTITUTION", "Constitution", "CONST", "Const."]),
        ],
        [{"ORTH": "Rules"}, OF, COURT],
        titled_words("old code"),
        titled_words("philippine civil code"),
        titled_words("civil code of 1950"),
        titled_words("civil code of 1889"),
        titled_words("civil code"),
        titled_words("civil code") + OF_THE_PH_,
        lower_words("revised administrative code of 1987"),
        lower_words("revised administrative code of 1917"),
        lower_words("revised administrative code"),
        lower_words("administrative code of 1987"),
        lower_words("administrative code of 1917"),
        lower_words("administrative code of 1916"),
        lower_words("administrative code"),
        lower_words("admin. code"),
        lower_words("admin code"),
        lower_words("adm. code"),
        lower_words("adm code"),
    ],
)



import itertools
from enum import Enum
from typing import Any

from lextok.rules._pattern import VS, Label, Rule, _orth_in, _re
from lextok.rules.abbreviations import org_options, ph_options

extras = " ".join(org_options)
extras += " " + " ".join(ph_options)
extras += " . , et al. et al the Jr Jr. Sr Sr. III IV Partnership Dev't"
misc = _orth_in(extras.split()) | {"OP": "*"}
opt_acronym = _re("\\([A-Z]+\\)") | {"OP": "?"}


class CaseName(Enum):
    ent = {
        "ENT_TYPE": {
            "IN": [
                Label.ORG.name,
                Label.PERSON.name,
                Label.GovtDivision.name,
                Label.JuridicalEntity.name,
            ]
        },
        "OP": "+",
    }
    pos = {"POS": {"IN": ["PROPN", "ADP", "DET", "CCONJ"]}, "OP": "+"}
    upper = {"IS_UPPER": True, "OP": "{1,6}"}
    title = {"IS_TITLE": True, "OP": "{1,6}"}

    @staticmethod
    def _vs(first: dict[str, Any], second: dict[str, Any]) -> list[dict[str, Any]]:
        return [first, opt_acronym, misc] + [VS] + [second, opt_acronym, misc]

    @classmethod
    def permute_patterns(cls) -> list[list[dict[str, Any]]]:
        return [
            cls._vs(a, b) for a, b in itertools.permutations((m.value for m in cls), 2)
        ]


casename = Rule(label=Label.CaseName, patterns=CaseName.permute_patterns())



from enum import Enum
from typing import Any

from lextok.rules._pattern import CM, OPT_NUMS, Label, Rule, _multi

dig = {"IS_DIGIT": True}
pages = {"ORTH": {"IN": ["at", "p", "p.", ","]}, "OP": "?"}
covered = {"TEXT": {"REGEX": "^\\(\\d+\\)$"}, "OP": "?"}
publisher_short = {"ORTH": {"IN": ["SCRA", "Phil", "Phil.", "OG", "O.G."]}}
publisher_words_phil = [
    {"ORTH": {"IN": ["Phil", "Phil."]}},
    {"ORTH": {"IN": ["Rep", "Rep.", "Reports"]}},
]
publisher_words_og = [
    {"ORTH": {"IN": ["Off", "Off."]}},
    {"ORTH": {"IN": ["Gaz", "Gazette"]}},
]

generic_start: list[dict[str, Any]] = [dig, publisher_short, pages]
generic_start_phil_words: list[dict[str, Any]] = [dig] + publisher_words_phil + [pages]
generic_start_og_words: list[dict[str, Any]] = [dig] + publisher_words_og + [pages]
special_volumes: list[dict[str, Any]] = [
    {"ORTH": {"IN": ["258-A", "290-A", "8-A"]}},  # Dashed letter vs. digit
    publisher_short,
    dig,
]
connected_comma_pages: list[dict[str, Any]] = [
    dig,
    publisher_short,
    {"TEXT": {"REGEX": "\\d+[,-]\\d+"}},  # 21 Phil 124,125
]
og_legacy: list[dict[str, Any]] = [
    dig,
    covered,
    {"ORTH": {"IN": ["OG", "O.G."]}},
    covered,
]


reporter_nums = Rule(
    label=Label.ReporterNum,
    patterns=[
        generic_start + [dig],
        generic_start + [{"TEXT": {"REGEX": "\\d+[,-]\\d+"}}],
        generic_start_phil_words + [dig],
        generic_start_phil_words + [{"TEXT": {"REGEX": "\\d+[,-]\\d+"}}],
        generic_start_og_words + [dig],
        generic_start_og_words + [{"TEXT": {"REGEX": "\\d+[,-]\\d+"}}],
        connected_comma_pages,
        special_volumes,
        og_legacy + [dig],
        og_legacy + [{"LIKE_NUM": True}],  # e.g. fourth, fifth
    ],
)


stat = {
    "ENT_TYPE": {
        "IN": [
            Label.StatuteNamed.name,
            Label.StatuteNum.name,
            Label.DocketNum.name,
        ]
    }
}

opt_comma_plus = {"ORTH": {"IN": ["of", "the", ","]}, "OP": "*"}

linker = {"ORTH": {"IN": ["or", "and", ",", "&"]}, "OP": "+"}

multi_linked_statute_pattern = _multi([linker, stat], 10)  # type: ignore
for linked_list in multi_linked_statute_pattern:
    linked_list.insert(0, stat)
as_amended: list[dict[str, Any]] = [
    {"ORTH": ",", "IS_PUNCT": True, "OP": "?"},
    {"LOWER": "as", "OP": "?"},
    {"LOWER": "amended"},
    {"LOWER": "by", "OP": "?"},
]

linked_statutes = Rule(
    label=Label.StatutoryLink,
    patterns=multi_linked_statute_pattern + [[stat] + as_amended + [stat]],
)
"""Connect statutory numbers together: `RA 141, RA 4124, RA 5325`"""

statutory_provisions = Rule(
    label=Label.StatutoryProvision,
    patterns=[
        [{"ENT_TYPE": Label.ProvisionNum.name, "OP": "+"}, opt_comma_plus, stat],
        [{"ENT_TYPE": Label.ProvisionNum.name, "OP": "+"}, opt_comma_plus, stat],
        [stat, opt_comma_plus, {"ENT_TYPE": Label.ProvisionNum.name, "OP": "+"}],
    ],
)
"""Connect statutes with provisions: `Art. 2, Civil Code`"""

decision_citations = Rule(
    label=Label.DecisionCitation,
    patterns=[
        [
            Label.DocketNum.node,
            OPT_NUMS,
            Label.DATE.node,
            CM | {"OP": "?"},
            Label.ReporterNum.node,
        ],
        [
            Label.DocketNum.node,
            OPT_NUMS,
            Label.DATE.node,
        ],
        [
            Label.ReporterNum.node,
            OPT_NUMS,
            CM | {"OP": "?"},
            Label.DATE.node,
        ],
    ],
)
"""Connect decision names, docket citations, date, and/or reporter numbers: `X v. Y, GR No. 3425, Jan. 1, 2000, 14 SCRA 14`"""



from typing import Any

from lextok.rules._pattern import OPT_NUMS, Label, Rule

up: dict[str, Any] = {"IS_UPPER": True, "OP": "+"}
tit: dict[str, Any] = {"IS_TITLE": True, "OP": "+"}
opt_tit: dict[str, Any] = {"IS_TITLE": True, "OP": "*"}
opt_cover: dict[str, Any] = {"TEXT": {"REGEX": "\\(.*\\)"}, "OP": "?"}
cov_num: list[dict[str, Any]] = [opt_cover, Label.GenericNum.node]
plus_date: list[dict[str, Any]] = [OPT_NUMS, Label.DATE.node]

doc_variants: list[list[dict[str, Any]]] = [
    [up, tit] + cov_num + plus_date,
    [up] + cov_num + plus_date,
    [tit] + cov_num + plus_date,
    [up, tit] + cov_num,
    [up] + cov_num,
    [tit] + cov_num,
    [up, tit, Label.GenericNum.node],
    [up, Label.GenericNum.node],
    [tit, Label.GenericNum.node],
    [Label.ORG.node, Label.GenericNum.node],
]
generic_documents = Rule(label=Label.Candidate, patterns=doc_variants)
"""Connect a generic serial number with a date, e.g. `Doc No. 414, Jan. 15, 2000`"""


irr_variants: list[list[dict[str, Any]]] = [
    [
        {"LOWER": "rules", "IS_TITLE": True},
        {"LOWER": "and", "IS_LOWER": True},
        {"LOWER": "regulations", "IS_TITLE": True},
        {"LOWER": "implementing", "IS_TITLE": True},
        {"IS_TITLE": True, "OP": "*"},
    ],
    [
        {"LOWER": "implementing", "IS_TITLE": True},
        {"LOWER": "rules", "IS_TITLE": True},
        {"LOWER": "and", "IS_LOWER": True},
        {"LOWER": "regulations", "IS_TITLE": True},
        {"ORTH": {"IN": ["of", "the"]}, "OP": "*"},
        {"IS_TITLE": True, "OP": "*"},
    ],
    [
        {"LOWER": {"IN": ["old", "new", "pre"]}, "IS_LOWER": True, "OP": "?"},
        {"ORTH": "-", "OP": "?"},
        {"SHAPE": "dddd", "OP": "?"},  # 1997, 1964, 2004
        {"IS_UPPER": True, "OP": "*"},  # NLRC, COMELEC, DARAB, Revised
        {"IS_TITLE": True, "OP": "*"},  # Revised
        {"LOWER": {"IN": ["rule", "rules"]}, "IS_TITLE": True},
        {"LOWER": {"IN": ["on", "of"]}},
        {"IS_TITLE": True, "OP": "*"},
    ],
    [
        {"SHAPE": "dddd", "OP": "?"},  # 1997, 1964, 2004
        {"IS_UPPER": True, "OP": "*"},  # NLRC
        {"IS_TITLE": True, "OP": "*"},  # Revised
        {"LOWER": "revised", "OP": "?"},
        {
            "LOWER": {"IN": ["omnibus", "uniform", "implementing", "rules", "interim"]},
            "IS_TITLE": True,
            "OP": "+",
        },
    ],
]
opt_connect: dict[str, Any] = {"ORTH": {"IN": ["of", "the", ","]}, "OP": "*"}

generic_irrs = Rule(label=Label.Candidate, patterns=irr_variants)


prov_irr = Rule(
    label=Label.CandidateProvision,
    patterns=[
        [{"ENT_TYPE": Label.ProvisionNum.name, "OP": "+"}, opt_connect] + i
        for i in irr_variants
    ]
    + [
        [{"ENT_TYPE": Label.ProvisionNum.name, "OP": "+"}, opt_connect] + d
        for d in doc_variants
    ],
)



from collections.abc import Iterator
from enum import Enum
from typing import Any

from lextok.rules._pattern import (
    CONNECTOR,
    NUM_SYM,
    DigitLists,
    Label,
    Rule,
    _orth_in,
    _re,
)
from lextok.rules.abbreviations import Prov


class ProvisionNum(Enum):
    """A statute's text is divided into (sometimes hierarchical) provisions.
    This structure combines both the adjective division, e.g. `Sec.`, `Section`, `Bk.`, etc.
    (which may have different casing and abbreviations) with presumed valid serial numbers,
    e.g. `1`, `1-b`, `III`, etc.
    """

    abbreviated_Sec1 = list(Prov.set_abbvs())
    abbreviated_sec1 = list(Prov.set_abbvs(cased="lower"))
    abbreviated_SEC1 = list(Prov.set_abbvs(cased="upper"))
    Section1 = list(Prov.set_fulls())
    section1 = list(Prov.set_fulls(cased="lower"))
    SECTION1 = list(Prov.set_fulls(cased="upper"))

    @classmethod
    def generate(cls) -> Iterator[list[dict[str, Any]]]:
        specific_digits = _re("(" + "|".join(DigitLists.generate_options()) + ")")
        for member in cls:
            for end in [{"POS": "NUM"}, specific_digits]:
                start = [CONNECTOR]  # start
                if member.name.startswith("abbreviated_"):
                    yield start + [_orth_in(member.value), {"ORTH": "."}, end]
                    yield start + [_orth_in([f"{v}." for v in member.value]), end]
                yield start + [_orth_in(member.value), end]


special_clauses: list[dict[str, Any]] = [
    {"ENT_TYPE": "ORDINAL"},
    {"LOWER": {"IN": ["whereas"]}},
    {"LOWER": {"IN": ["clause", "clauses"]}},
]
provision_nums = Rule(
    label=Label.ProvisionNum, patterns=list(ProvisionNum.generate()) + [special_clauses]
)


generic_nums = Rule(
    label=Label.GenericNum,
    patterns=[[NUM_SYM, {"POS": "NUM"}]],
)
